//
// Created by robin on 1/7/18.
//

#include <QDataStream>
#include "dns/QDnsHeader.h"

namespace QDns
{

    QDnsHeader::QDnsHeader()
            : m_isValid(true)
            , m_id(0)
            , m_flags(0)
            , m_queryCount(0)
            , m_answerCount(0)
            , m_authorityCount(0)
            , m_additionalCount(0)
    {

    }

    QDnsHeader QDnsHeader::parse(const QByteArray& data)
    {
        QDataStream dataStream(data);
        return QDnsHeader::parse(dataStream);
    }

    QDnsHeader QDnsHeader::parse(QDataStream& dataStream)
    {
        QDnsHeader header;

        if (dataStream.device()->bytesAvailable() < 12)
        {
            header.m_isValid = false;
        }
        else
        {
            dataStream >> header.m_id;
            dataStream >> header.m_flags;
            dataStream >> header.m_queryCount;
            dataStream >> header.m_answerCount;
            dataStream >> header.m_authorityCount;
            dataStream >> header.m_additionalCount;
        }

        return header;
    }

    QByteArray QDnsHeader::serialize() const
    {
        QByteArray byteArray;
        QDataStream dataStream(&byteArray, QIODevice::WriteOnly);
        serialize(dataStream);
        return byteArray;
    }

    void QDnsHeader::serialize(QDataStream& dataStream) const
    {
        dataStream << m_id;
        dataStream << m_flags;
        dataStream << m_queryCount;
        dataStream << m_answerCount;
        dataStream << m_authorityCount;
        dataStream << m_additionalCount;
    }

    bool QDnsHeader::isValid() const
    {
        return m_isValid;
    }

    quint16 QDnsHeader::getId() const
    {
        return m_id;
    }

    void QDnsHeader::setId(quint16 id)
    {
        m_id = id;
    }

    quint16 QDnsHeader::getFlags() const
    {
        return m_flags;
    }

    void QDnsHeader::setFlags(quint16 flags)
    {
        m_flags = flags;
    }

    void QDnsHeader::setFlagBits(quint8 bit, quint16 value, quint8 bitCount)
    {
        quint8 baseBit = getFlagBitIndex(bit);
        for (quint8 i = 0; i < bitCount; ++i) // TODO might be optimized
        {
            if ((value >> i) & 1)
            {
                m_flags |= 1UL << (baseBit + i);
            }
            else
            {
                m_flags &= ~(1UL << (baseBit + i));
            }
        }
    }

    quint16 QDnsHeader::getFlagBits(quint8 bit, quint8 bitCount) const
    {
        quint8 mask = 1;
        for (quint8 i = 1; i < bitCount; ++i)
        {
            mask += 1 << i;
        }
        auto value = (quint16) ((m_flags >> getFlagBitIndex(bit)) & mask);
        return value;
    }

    quint8 QDnsHeader::getFlagBitIndex(quint8 bit)
    {
        auto index = ((sizeof(m_flags) * 8) - bit - 1);
        return index;
    }

    bool QDnsHeader::isQuery() const
    {
        return (bool) getFlagBits(0, 1);
    }

    void QDnsHeader::setIsQuery(bool isQuery)
    {
        setFlagBits(0, (quint8) isQuery, 1);
    }

    quint8 QDnsHeader::getOpcode() const
    {
        return (quint8) getFlagBits(4, 4);
    }

    void QDnsHeader::setOpcode(quint8 opcode)
    {
        setFlagBits(4, opcode, 4);
    }

    bool QDnsHeader::isAuthoritative() const
    {
        return (bool) getFlagBits(5, 1);
    }

    void QDnsHeader::setIsAuthoritative(bool isAuthoritative)
    {
        setFlagBits(5, (quint8) isAuthoritative, 1);
    }

    bool QDnsHeader::isTruncated() const
    {
        return (bool) getFlagBits(6, 1);
    }

    void QDnsHeader::setIsTruncated(bool isTruncated)
    {
        setFlagBits(6, (quint8) isTruncated, 1);
    }

    bool QDnsHeader::isRecursionDesired() const
    {
        return (bool) getFlagBits(7, 1);
    }

    void QDnsHeader::setIsRecursionDesired(bool isRecursionDesired)
    {
        setFlagBits(7, (quint8) isRecursionDesired, 1);
    }

    bool QDnsHeader::isRecursionAvailable() const
    {
        return (bool) getFlagBits(8, 1);
    }

    void QDnsHeader::setIsRecursionAvailable(bool isRecursionAvailable)
    {
        setFlagBits(8, (quint8) isRecursionAvailable, 1);
    }

    bool QDnsHeader::isAnswerAuthenticated() const
    {
        return (bool) getFlagBits(10, 1);
    }

    void QDnsHeader::setIsAnswerAuthenticated(bool isAnswerAuthenticated)
    {
        setFlagBits(10, (quint8) isAnswerAuthenticated, 1);
    }

    bool QDnsHeader::isNonAuthenticatedAcceptable() const
    {
        return (bool) getFlagBits(11, 1);
    }

    void QDnsHeader::setNonAuthenticatedAcceptable(bool isNonAuthenticatedAcceptable)
    {
        setFlagBits(11, (quint8) isNonAuthenticatedAcceptable, 1);
    }

    quint8 QDnsHeader::getReplyCode() const
    {
        return (quint8) getFlagBits(15, 4);
    }

    void QDnsHeader::setReplyCode(quint8 replyCode)
    {
        setFlagBits(15, (quint8) replyCode, 4);
    }

    quint16 QDnsHeader::getQueryCount() const
    {
        return m_queryCount;
    }

    void QDnsHeader::setQueryCount(quint16 questionCount)
    {
        m_queryCount = questionCount;
    }

    quint16 QDnsHeader::getAnswerCount() const
    {
        return m_answerCount;
    }

    void QDnsHeader::setAnswerCount(quint16 answerCount)
    {
        m_answerCount = answerCount;
    }

    quint16 QDnsHeader::getAuthorityCount() const
    {
        return m_authorityCount;
    }

    void QDnsHeader::setAuthorityCount(quint16 authorityCount)
    {
        m_authorityCount = authorityCount;
    }

    quint16 QDnsHeader::getAdditionalCount() const
    {
        return m_additionalCount;
    }

    void QDnsHeader::setAdditionalCount(quint16 additionalCount)
    {
        m_additionalCount = additionalCount;
    }

}